#include <g726.h>
#include <re.h>
#include <baresip.h>


#define DEBUG_MODULE "g726-32"
#define DEBUG_LEVEL 5
#include <re_dbg.h>

struct aucodec_st {
	struct aucodec *ac;  /* inheritance */
	g726_state_t enc, dec;
};

static struct aucodec *ac_g726_32;

static void g726_32_destructor(void *arg)
{
	struct aucodec_st *st = arg;

	mem_deref(st->ac);
}


static int alloc(struct aucodec_st **stp, struct aucodec *ac,
		 struct aucodec_prm *encp, struct aucodec_prm *decp,
		 const char *fmtp)
{
	struct aucodec_st *st;
	int err = 0;

	(void)encp;
	(void)decp;
	(void)fmtp;

	st = mem_zalloc(sizeof(*st), g726_32_destructor);
	if (!st)
		return ENOMEM;

	st->ac = mem_ref(ac);

	if (!g726_init(&st->enc, 32000, G726_PACKING_LEFT)) {
		DEBUG_WARNING("g726_init() encoder failed\n");
		err = EPROTO;
		goto out;
	}

	if (!g726_init(&st->dec, 32000, G726_PACKING_LEFT)) {
		DEBUG_WARNING("g726_init() decoder failed\n");
		err = EPROTO;
		goto out;
	}

 out:
	if (err)
		mem_deref(st);
	else
		*stp = st;

	return err;
}


static int encode(struct aucodec_st *st, struct mbuf *dst, struct mbuf *src)
{
	int16_t *sig;
	uint8_t *byte;
	size_t nsamp, bytes;

	nsamp = mbuf_get_left(src) / 2;

	if (mbuf_get_space(dst) < nsamp) {
		int err = mbuf_resize(dst, dst->size + nsamp);
		if (err)
			return err;
	}

	sig = (int16_t *)mbuf_buf(src);
	byte = mbuf_buf(dst);

	bytes = g726_encode(&st->enc, byte, sig, nsamp);

	mbuf_advance(src, nsamp*2);
	dst->end += bytes;

	return 0;
}


static int decode(struct aucodec_st *st, struct mbuf *dst, struct mbuf *src)
{
	int16_t *sig;
	uint8_t *byte;
	size_t nsamp, samples;

	nsamp = mbuf_get_left(src) * 2;
	if (!nsamp)
		return 0;

	if (mbuf_get_space(dst) < 4*nsamp) {
		int err = mbuf_resize(dst, dst->size + 4*nsamp);
		if (err)
			return err;
	}

	byte = mbuf_buf(src);
	sig = (int16_t *)mbuf_buf(dst);

	samples = g726_decode(&st->dec, sig, byte, nsamp/2);

	mbuf_advance(src, nsamp/2);
	dst->end += samples*2;

	return 0;
}


static int module_init(void)
{
	return aucodec_register(&ac_g726_32, NULL, "G726-32", 8000, 1, NULL,
				alloc, encode, decode, NULL);
}


static int module_close(void)
{
	ac_g726_32 = mem_deref(ac_g726_32);
	return 0;
}


EXPORT_SYM const struct mod_export DECL_EXPORTS(g726_32) = {
	"g726_32",
	"codec",
	module_init,
	module_close
};
